"use strict";

const React = require("react");
const classnames = require("classnames");
const assert = require("assert");
const defaultValue = require("default-value");
const dateFns = require("date-fns");
const debounce = require("debounce");

const Frame = require("./frame");
const MainView = require("./main-view");
const StatusBall = require("./status-ball");
const {Form, Input, SubmitButton} = require("./form");

const humanReadableStatus = require("../human-readable-status");
const useStore = require("../plumbing/use-store");
const useModule = require("../plumbing/use-module");
const useTimeout = require("../use-timeout");
const useLockableScroll = require("../use-lockable-scroll");

/* FIXME: Input persistence, both 1) contents when switching tabs, and 2) focus */

function Status({friend}) {
	if (friend.statusVisible === true) {
		return (
			<div className="status">
				<StatusBall status={friend.status} statusVisible={friend.statusVisible} />
				{humanReadableStatus(friend.status)}
			</div>
		);
	} else {
		return null;
	}
}

function StatusMessage({friend}) {
	let prefix = (friend.statusVisible === true) ? "• " : "";

	return (
		<div className="statusMessage">
			{(friend.statusMessage != null)
				? `${prefix}${friend.statusMessage}`
				: undefined}
		</div>
	);
}

function ConversationFriendInformation({friend}) {
	return (
		<div className="friendInformation">
			<div className="name">
				<div className="displayName">
					{defaultValue(friend.displayName, friend.username)}
				</div>
				<div className="username">
					{friend.username}
				</div>
			</div>
			<div className="statusDisplay">
				<Status friend={friend} />
				<StatusMessage friend={friend} />
			</div>
		</div>
	);
}

function ConversationFriendAction({className, children}) {
	return (
		<div className={classnames("action", className)}>
			{children}
		</div>
	);
}

function ConversationFriendActions() {
	return (
		<div className="friendActions">
			<ConversationFriendAction className="sendPicture">
				Send picture
			</ConversationFriendAction>
			<ConversationFriendAction className="sendFile">
				Send file
			</ConversationFriendAction>
			<ConversationFriendAction className="remove">
				Remove friend
			</ConversationFriendAction>
		</div>
	);
}

function ConversationViewHeader({friend}) {
	return (
		<MainView.Header>
			<ConversationFriendInformation friend={friend} />
			<ConversationFriendActions />
		</MainView.Header>
	);
}

function MessageSpinner() {
	return <div className="messageSpinner" />;
}

function Message({message}) {
	let sendPeriodElapsed = useTimeout(2000, message.timestamp);

	return (
		<div className="message sent">
			<div className="body">
				{(message.dbStored === false && sendPeriodElapsed)
					? <MessageSpinner />
					: null}
				{message.body}
			</div>
		</div>
	);
}

function MessageGroup({messageGroup}) {
	let {messages} = useModule("messages").state;
	let {friends} = useModule("friends").state;
	
	let senderName;

	if (messageGroup.direction === "sent") {
		senderName = "You";
	} else {
		let friend = friends.get(messageGroup.sender);
		senderName = defaultValue(friend.displayName, friend.username);
	}

	/* FIXME: Why is there a hard-coded `sent` class name? */
	return (
		<div className="messageGroup sent">
			<div className="metadata">
				<div className="timestamp">
					{dateFns.format(messageGroup.firstTimestamp, "HH:mm")}
				</div>
				<div className="sender">
					{senderName}
				</div>
			</div>
			{messageGroup.messageIds.map((messageId) => {
				let message = messages.get(messageId);

				return <Message key={message.id} message={message} />;
			})}
		</div>
	);
}

function ConversationViewContents({messageGroups, friend}) {
	let {dispatch} = useModule("friends");
	let {getState} = useStore();
	let {scrollTop, scrollLocked} = friend;

	/* MARKER: Get rid of local locked state, it's totally useless when dealing with views anyway */
	/* MARKER: Move debounce to within lockable-scroll abstraction, based on scrollKey for invalidation */
	let [locked, setRef] = useLockableScroll({
		scrollTop: scrollTop,
		locked: scrollLocked,
		scrollKey: `${friend.messageIds.size}::${friend.username}`,
		onScroll: debounce((locked, scrollTop) => {
			/* HACK: The below check is because the `debounce` will defer a call to this function when switching a tab, and so this function will be called and cause an erroneous scroll-set *after* we've already ended up on another user's tab. */
			let {selectedTabType, selectedTab} = getState();

			if (selectedTabType === "conversation" && selectedTab === friend.username) {
				dispatch({
					type: "scrolled",
					username: friend.username,
					locked: locked,
					scrollTop: scrollTop
				});
			}
		}, 100)
	});

	return (
		<MainView.Contents ref={setRef} className={classnames("conversation", {locked: friend.scrollLocked})}>
			{messageGroups.map((messageGroup) => {
				return <MessageGroup key={messageGroup.id} messageGroup={messageGroup} />;
			})}
		</MainView.Contents>
	);
}

function MessageInput({friend}) {
	return (
		<Frame className="inputBox">
			<Form module="uiMessageInput" stateKey={friend.username}>
				{/* <TextArea name="body" /> */}
				<Input name="body" autoFocus />
				<SubmitButton>Send</SubmitButton>
			</Form>
		</Frame>
	);
}

module.exports = function ConversationView() {
	let {friends} = useModule("friends").state;
	let {state} = useStore();
	let {selectedTab, selectedTabType} = state;

	assert(selectedTabType === "conversation");

	let friend = friends.get(selectedTab);
	assert(friend != null);

	return (
		<div className="conversationView">
			<MainView.Container>
				<ConversationViewHeader friend={friend} />
				<ConversationViewContents friend={friend} messageGroups={friend.messageGroups.toArray()} />
			</MainView.Container>
			<MessageInput friend={friend} />
		</div>
	);
};