"use strict";

const React = require("react");
const dotty = require("dotty");
const classnames = require("classnames");
const formSerialize = require("form-serialize");

const Spinner = require("./spinner");
const useModule = require("../plumbing/use-module");
const types = require("../plumbing/types");

function getFormState(state, key) {
	if (state.formState.has(key)) {
		return state.formState.get(key);
	} else {
		return types.UiFormState();
	}
}

let FormContext = React.createContext();

function Form({module, statePath, className, children, stateKey}) {
	let {state, dispatch} = useModule(module);
	let [formRef, setFormRef] = React.useState();
	let formState = getFormState(state, stateKey);

	function handleSubmit(event) {
		event.preventDefault();

		if (formRef != null) {
			let formData = formSerialize(formRef, {hash: true});

			dispatch({
				type: "submit",
				formData: formData,
				key: stateKey
			});
		}
	}

	React.useEffect(() => {
		if (formRef != null) {
			for (let element of formRef.querySelectorAll("input")) {
				element.value = "";
			}
		}
	}, [formState.resetCounter]);

	let contextValue = {
		formState: formState,
		formStateKey: stateKey,
		submit: handleSubmit,
		handleBlur: function (event) {
			dispatch({
				type: "blur",
				key: stateKey, /* FIXME: Abstract this out everywhere */
				field: event.target.name
			});
		},
		handleFocus: function (event) {
			dispatch({
				type: "focus",
				key: stateKey,
				field: event.target.name
			});
		},
		handleUnmount: function () {
			
		}
	}

	return (
		<div className="form">
			<form ref={setFormRef} className={className}>
				<FormContext.Provider value={contextValue}>
					{children}
				</FormContext.Provider>
			</form>

			{formState.errorMessage != null
				? <div className="error form">{formState.errorMessage}</div>
				: null}
		</div>
	);
}

function Input({type="text", name, placeholder, className, autoFocus}) {
	let {formState, formStateKey, handleBlur, handleFocus} = React.useContext(FormContext);
	let [inputRef, setInputRef] = React.useState();

	let isInvalid = formState.invalidFields.has(name);

	/* The below restores focus after re-rendering the form (eg. because the resetCounter was triggered after a form submission) */
	React.useEffect(() => {
		let needsFocus = inputRef != null && formState.wasFocused === name;

		if (needsFocus) {
			inputRef.focus();
		}
	}, [inputRef, formState.wasFocused, formState.resetCounter]);

	React.useEffect(() => {
		/* MARKER: Re-focus after form submission */
		if (autoFocus === true && inputRef != null) {
			inputRef.focus();
		}
	}, [name, formStateKey, inputRef]);

	return <input ref={setInputRef} onBlur={handleBlur} onFocus={handleFocus} disabled={formState.processing} className={classnames(className, {invalid: isInvalid})} type={type} name={name} placeholder={placeholder} />;
}

function TextArea({name, placeholder, className}) {
	let {formState} = React.useContext(FormContext);
	
	let isInvalid = formState.invalidFields.has(name);

	return <textarea disabled={formState.processing} className={classnames(className, {invalid: isInvalid})} name={name} placeholder={placeholder} />;
}

function SubmitButton({children}) {
	let {submit, formState} = React.useContext(FormContext);

	return <button disabled={formState.processing} type="submit" onClick={submit}>{children}</button>;
}

function FormSpinner() {
	let {formState} = React.useContext(FormContext);

	if (formState.processing) {
		return <Spinner />;
	} else {
		return null;
	}
}

module.exports = {Form, Input, TextArea, SubmitButton, FormSpinner};