"use strict";

const immutable = require("immutable");

const types = require("../plumbing/types");

function removeFriendRequest(username) {
	return function (state) {
		return state.updateIn(["friendRequests"], (requests) => requests.delete(username));
	}
}

module.exports = {
	name: "friendRequests",
	state: {
		friendRequests: immutable.Map()
	},
	stateTransformations: {
		received: (state, action) => {
			let {request} = action;

			return state.updateIn(["friendRequests"], (requests) => requests.set(request.username, request));
		},
		withdrawn: (state, action) => {
			let {username} = action;

			return state.update(removeFriendRequest(username));
		},
		accept: (state, action) => {
			let {username} = action;

			return state.setIn(["friendRequests", username, "beingAccepted"], true);
		},
		accepted: (state, action) => {
			let {username} = action;

			/* NOTE: This *only* handles the `friendRequests` entry, not the addition to the friends list. That's a separate `friendAdded` action. */

			return state.update(removeFriendRequest(username));
		},
		reject: (state, action) => {
			let {username} = action;

			return state.setIn(["friendRequests", username, "beingRejected"], true);
		},
		rejected: (state, action) => {
			let {username} = action;

			return state.update(removeFriendRequest(username));
		}
	},
	logicHandlers: {
		accept: {
			errorType: "acceptFailed",
			process: ({action, api}) => {
				return api.acceptFriendRequest({
					username: action.username
				});
			},
			success: ({result, action}) => {
				return [{
					type: "accepted",
					username: action.username
				}, {
					type: "@friends_added",
					friend: types.Friend(result.friend)
				}];
			},
			failure: ({error}) => {
				return {
					errorMessage: error.message
				};
			}
		},
		reject: {
			errorType: "rejectFailed",
			process: ({action, api}) => {
				return api.rejectFriendRequest({
					username: action.username
				});
			},
			success: ({action}) => {
				return {
					type: "rejected",
					username: action.username
				};
			},
			failure: ({error}) => {
				return {
					errorMessage: error.message
				};
			}
		}
	}
};