"use strict";

const immutable = require("immutable");
const nanoid = require("nanoid");

const types = require("../plumbing/types");

module.exports = {
	name: "friends",
	state: {
		friends: immutable.Map(),
	},
	stateTransformations: {
		added: (state, action) => {
			let {friend} = action;

			return state.updateIn(["friends"], (friends) => friends.set(friend.username, friend));
		},
		removed: (state, action) => {
			let {username} = action;

			return state.updateIn(["friends"], (friends) => friends.delete(username));
		},
		startedTyping: (state, action) => {
			return state.setIn(["friends", username, "isTyping"], true);
		},
		pausedTyping: (state, action) => {
			return state.setIn(["friends", username, "pausedTyping"], true);
		},
		resumedTyping: (state, action) => {
			return state.setIn(["friends", username, "pausedTyping"], false);
		},
		stoppedTyping: (state, action) => {
			return state.setIn(["friends", username, "isTyping"], false);
		},
		changedStatus: (state, action) => {
			let {username, status} = action;

			/* TODO: Figure out a way to express the branching below in a functional, expression-based manner. */
			let state_ = state.setIn(["friends", username, "status"], status);

			if (status === "offline") {
				return state_.mergeIn(["friends", username], {
					isTyping: false,
					pausedTyping: false
				});
			} else {
				return state_;
			}
		},
		changedStatusMessage: (state, action) => {
			let {statusMessage} = action;

			return state.setIn(["friends", username, "statusMessage"], statusMessage);
		},
		changedDisplayName: (state, action) => {
			let {username, displayName} = action;

			return state.setIn(["friends", username, "displayName"], displayName);
		},
		addedMessageToConversation: (state, action) => {
			let {username, message} = action;

			let messageGroups = state.getIn(["friends", username, "messageGroups"]);
			let messageGroupCount = messageGroups.size;
			let lastMessageGroup = messageGroups.last();
			let shouldCreateNewMessageGroup = (lastMessageGroup == null) || (lastMessageGroup.direction !== message.direction);

			let state_ = state.updateIn(["friends", username, "messageIds"], (messageIds) => {
				return messageIds.push(message.id);
			});

			if (shouldCreateNewMessageGroup) {
				return state_.updateIn(["friends", username, "messageGroups"], (messageGroups) => {
					return messageGroups.push(types.MessageGroup({
						id: nanoid(),
						direction: message.direction,
						sender: message.sender,
						recipient: message.recipient,
						firstTimestamp: message.timestamp,
						messageIds: immutable.List([message.id])
					}));
				});
			} else {
				return state_.updateIn(["friends", username, "messageGroups", messageGroupCount - 1, "messageIds"], (messageIds) => {
					return messageIds.push(message.id);
				});
			}
		},
		scrolled: (state, action) => {
			let {username, locked, scrollTop} = action;

			return state.mergeIn(["friends", username], {
				scrollTop: scrollTop,
				scrollLocked: locked
			});
		}
	}
};