"use strict";

const immutable = require("immutable");

const types = require("../plumbing/types");

/* Additional exports:
- creationFailed
- deletionFailed

IDEA: Move the API logic away from UI modules, and into 'core modules' like this one. Then export the success/failure actions/events from the core modules, so that the UI modules can respond to them. This centralizes all logic, so that it can be called from multiple places with a single implementation.

Caveat: For UI elements to work correctly (eg. not erroneously point out error conditions that are no longer applicable), it needs to be possible to not only tie results to requests, but also to know what the latest operation of a particular type was.
*/

module.exports = {
	name: "invites",
	state: {
		invites: immutable.Map() // string [id] -> Invite
	},
	stateTransformations: {
		created: (state, action) => {
			let {invite} = action;

			return state.updateIn(["invites"], (invites) => {
				return invites.set(invite.id, invite);
			});
		},
		consumed: (state, action) => {
			let {inviteId} = action;

			return state.setIn(["invites", inviteId, "consumed"], true);
		},
		delete: (state, action) => {
			let {inviteId} = action;

			return state.setIn(["invites", inviteId, "beingDeleted"], true);
		},
		deleted: (state, action) => {
			let {inviteId} = action;
			
			return state.updateIn(["invites"], (invites) => {
				return invites.delete(inviteId);
			});
		},
		deleteFailed: (state, action) => {
			/* FIXME: Don't just log to the console, but show feedback to the user, somehow... */
			console.warn(`Invite delete failed: ID ${action.sourceAction.inviteId}`);
			return state;
		}
	},
	logicHandlers: {
		delete: {
			errorType: "deleteFailed",
			validate: ({allow, reject, state, action}) => {
				if (state.invites.get(action.inviteId).consumed === true) {
					reject({
						errorMessage: "Cannot delete invite code that has already been used"
					});
				} else {
					allow();
				}
			},
			process: ({action, api}) => {
				return api.deleteInvite({id: action.inviteId});
			},
			success: ({action}) => {
				return {
					type: "deleted",
					inviteId: action.inviteId
				};
			},
			failure: ({error}) => {
				/* FIXME: Error filtering */
				return {
					errorMessage: error.message
				};
			}
		}
	}
};