"use strict";

const immutable = require("immutable");
const defaultValue = require("default-value");

const types = require("../../plumbing/types");
const mapDispatch = require("../../map-dispatch");
const withArgumentActionMapper = require("../../actions/with-argument-action-mapper");
const mapActions = require("../../actions/map-actions");

function dispatchWithKey(dispatch, key) {
	return mapDispatch(dispatch, (action) => {
		return Object.assign({key: key}, action);
	});
}

function defaultFormState(formState) {
	if (formState != null) {
		return formState;
	} else {
		return types.UiFormState();
	}
}

function keyMapper(key) {
	return function(action) {
		return Object.assign({key: key}, action);
	};
}

module.exports = function ({validate, transform, process, success}) {
	return {
		state: {
			formState: immutable.Map() // string [key] => UiFormState
		},
		stateTransformations: {
			processing: (state, action) => {
				let {key} = action;

				return state.updateIn(["formState", key], (formState) => {
					let formState_ = defaultFormState(formState);

					return formState_.merge({
						processing: true,
						errorMessage: null,
						invalidFields: formState_.invalidFields.clear()
					});
				});
			},
			error: (state, action) => {
				let {errorMessage, invalidFields, key} = action;
	
				return state.updateIn(["formState", key], (formState) => {
					let formState_ = defaultFormState(formState);

					return formState_.merge({
						processing: false,
						errorMessage: errorMessage,
						invalidFields: immutable.Set(invalidFields)
					});
				});
			},
			complete: (state, action) => {
				let {key} = action;

				return state.updateIn(["formState", key], (formState) => {
					let formState_ = defaultFormState(formState);

					return formState_.merge({
						processing: false,
						resetCounter: formState_.resetCounter + 1
					});
				});
			},
			blur: (state, action) => {
				let {key} = action;

				/* FIXME: Abstract out the below wrapping */
				return state.updateIn(["formState", key], (formState) => {
					let formState_ = defaultFormState(formState);

					return formState_.merge({
						wasFocused: null
					});
				});
			},
			focus: (state, action) => {
				let {key, field} = action;

				return state.updateIn(["formState", key], (formState) => {
					let formState_ = defaultFormState(formState);

					return formState_.merge({
						wasFocused: field
					});
				});
			}
		},
		logicHandlers: {
			submit: {
				errorType: "error",
				validate: ({globalState, state, action, allow, reject}) => {
					let validateFunc = defaultValue(validate, transform);
					
					validateFunc({
						globalState: globalState,
						state: state,
						formData: action.formData,
						allow: withArgumentActionMapper(allow, keyMapper(action.key), {outputSingleDispatch: true}),
						reject: withArgumentActionMapper(reject, keyMapper(action.key), {outputSingleDispatch: true}),
						key: action.key
					});
				},
				process: ({action, api, dispatch}) => {
					let mappedDispatch = dispatchWithKey(dispatch, action.key);
					mappedDispatch("processing"); /* FIXME: Is this necessary? Can't we just rename the `processing` transformation to `process` instead? */
					
					return process({
						api: api,
						formData: action.formData
					});
				},
				success: ({action, result}) => {
					let successActions = success({
						result: result
					});

					let defaultSuccessActions = [{
						type: "complete"
					}];

					let allSuccessActions;

					/* FIXME: Verify that this logic works for single (non-array) returned actions */
					/* FIXME: Fail early on non-existent scoped actions, so that the error message can show the scoped name and module name instead of the globalized name. */
					if (successActions != null) {
						allSuccessActions = defaultSuccessActions.concat(successActions);
					} else {
						allSuccessActions = defaultSuccessActions;
					}

					return mapActions(allSuccessActions, keyMapper(action.key));
				},
				/* FIXME: Why did the below appear to work when the signature was still (error)=> ? */
				failure: ({error, action}) => {
					/* FIXME: Error filtering to distinguish bugs from end-user errors? */
					console.error(error.stack);
					return {
						key: action.key,
						errorMessage: error.message
					};
				}
			}
		}
	};
}