"use strict";

const Promise = require("bluebird");
const reduxLogic = require("redux-logic");
const assureArray = require("assure-array");
const assert = require("assert");

module.exports = function createLogic({type, validate, transform, process, success, failure, errorType, actionMapper = (action) => action, actionUnmapper = (action) => action, dataMapper = (data) => data}) {
	assert(errorType != null);

	/* FIXME: Add support for never-ending processes (ie. no done() called ever) */
	/* FIXME: Implement a `transform` wrapper; at least `actionMapper` support, possible other changes also. */
	/* QUESTION: Is dataMapper(unmapData(data)) the sensible order, or should it be the other way around? */
	let validateFunc, processFunc;

	function unmapData(data) {
		return Object.assign({}, data, {
			action: actionUnmapper(data.action)
		});
	}

	if (validate != null) {
		validateFunc = function (data, allow, reject) {
			let originalAction = data.action;

			let allowHandler = function (allowAction) {
				if (allowAction != null) {
					allow(actionMapper(allowAction));
				} else {
					allow(originalAction);
				}
			};

			let rejectHandler = function (errorData) {
				if (typeof errorData === "string") {
					reject(actionMapper({
						type: errorData,
						sourceAction: originalAction
					}));
				} else {
					reject(actionMapper(Object.assign({
						type: errorType,
						error: true,
						sourceAction: originalAction
					}, errorData)));
				}
			}

			let props = Object.assign({
				allow: allowHandler,
				reject: rejectHandler
			}, dataMapper(unmapData(data)));

			validate(props);
		}
	}

	if (process != null) {
		processFunc = function (data, dispatch, done) {
			let originalAction = data.action;

			function dispatchWrapper(action) {
				if (typeof action === "string") {
					dispatch(actionMapper({type: action}));
				} else {
					dispatch(actionMapper(action));
				}
			}

			function dispatchAll(items, mapper = (item) => item) {
				assureArray(items).forEach((item) => {
					dispatchWrapper(mapper(item));
				});
			}

			return Promise.try(() => {
				let props = Object.assign({
					dispatch: dispatchWrapper
				}, dataMapper(unmapData(data)));

				return process(props);
			}).then((result) => {
				let successAction = success({
					result: result,
					action: originalAction
				});
				
				if (successAction != null) {
					dispatchAll(successAction);
				}
			}).catch((error) => {
				// console.error(error.message);

				/* FIXME: Some sort of flag on error objects to distinguish between bugs and end-user errors? */
				let errorData = failure({
					error: error,
					action: originalAction
				});

				if (errorData != null) {
					dispatchAll(errorData, (item) => {
						return Object.assign({
							type: errorType,
							error: true,
							sourceAction: originalAction
						}, item);
					});
				}
			}).finally(() => {
				done();
			});
		}
	}

	return reduxLogic.createLogic({
		type: type,
		validate: validateFunc,
		transform: transform,
		process: processFunc
	});
};