"use strict";

const immutable = require("immutable");

function addMessage(friendUsername, message) {
	return function (state) {
		return state
			.updateIn(["messages"], (messages) => messages.set(message.id, message))
			.updateIn(["friends", friendUsername, "messageIds"], (ids) => ids.push(message.id));
	}
}

function markMessageDbStored(messageId) {
	return function (state) {
		return state.setIn(["messages", messageId, "dbStored"], true);
	}
}

function removeFriendRequest(username) {
	return function (state) {
		return state.updateIn(["friendRequests"], (requests) => requests.delete(username));
	}
}

function clearFriendState(username) {
	return function (state) {
		return state.mergeIn(["friends", username], {
			isTyping: false,
			pausedTyping: false
		});
	}
}

function validateStatus(status) {
	if (!["online", "away", "offline"].includes(status)) {
		throw new Error(`Invalid status: ${status}`);
	} else {
		return status;
	}
}

function mapFromArray(sourceArray, mapper) {
	return immutable.Map(sourceArray.map((item) => {
		return mapper(item);
	}));
}

module.exports = {
	selectedConversationTab: (state, action) => {
		let {username} = action;

		return state.merge({
			selectedTabType: "conversation",
			selectedTab: username
		});
	},
	selectedAddFriendTab: (state, action) => {
		return state.merge({
			selectedTabType: "meta",
			selectedTab: "addFriend"
		});
	},
	connected: (state, action) => {
		let {username, displayName, statusMessage, status, friends} = action;

		/* FIXME: Verify whether seeding with initial data also works for module state. */
		/* TODO: Initialize Invitations, FriendRequests, etc. */
		
		let friendMap = mapFromArray(friends, (friend) => [friend.username, friend]);

		return state
			.merge({
				connected: true,
				username: username,
				displayName: displayName,
				status: validateStatus(status),
				statusMessage: statusMessage
			})
			.setIn(["modules", "friends", "friends"], friendMap);
	},
	disconnected: (state, action) => {
		return state.merge({
			connected: false
		});
	},
	userChangedStatus: (state, action) => {
		let {status} = action;

		return state.set("status", validateStatus(status));
	},
	userChangedStatusMessage: (state, action) => {
		let {statusMessage} = action;

		return state.set("statusMessage", statusMessage);
	},
	userChangedDisplayName: (state, action) => {
		let {displayName} = action;

		return state.merge({
			displayName: displayName,
			displayNameDBStored: false
		});
	},
	userChangedDisplayNameDBStored: (state, action) => {
		return state.set("displayNameDBStored", true);
	},
	userStartedTyping: (state, action) => {
		let {recipient} = action;

		return state.updateIn(["isTypingTo"], (friends) => {
			return friends.add(recipient);
		});
	},
	userPausedTyping: (state, action) => {
		let {recipient} = action;

		return state.updateIn(["pausedTypingTo"], (friends) => {
			return friends.add(recipient);
		});
	},
	userResumedTyping: (state, action) => {
		let {recipient} = action;

		return state.updateIn(["pausedTypingTo"], (friends) => {
			return friends.delete(recipient);
		});
	},
	userStoppedTyping: (state, action) => {
		let {recipient} = action;

		return state.updateIn(["isTypingTo"], (friends) => {
			return friends.delete(recipient);
		});
	},
	friendChangedStatus: (state, action) => {
		let {username, status} = action;

		/* TODO: Figure out a way to express the branching below in a functional, expression-based manner. */
		let state_ = state.setIn(["friends", username, "status"], status);

		if (status === "offline") {
			return state_.update(clearFriendState(username));
		} else {
			return state_;
		}
	},
	friendChangedStatusMessage: (state, action) => {
		let {statusMessage} = action;

		return state.setIn(["friends", username, "statusMessage"], statusMessage);
	},
	friendChangedDisplayName: (state, action) => {
		let {username, displayName} = action;

		return state.setIn(["friends", username, "displayName"], displayName);
	},
	friendStartedTyping: (state, action) => {
		return state.setIn(["friends", username, "isTyping"], true);
	},
	friendPausedTyping: (state, action) => {
		return state.setIn(["friends", username, "pausedTyping"], true);
	},
	friendResumedTyping: (state, action) => {
		return state.setIn(["friends", username, "pausedTyping"], false);
	},
	friendStoppedTyping: (state, action) => {
		return state.setIn(["friends", username, "isTyping"], false);
	},
	messageReceived: (state, action) => {
		let {message} = action;

		return state
			.update(addMessage(message.sender, message.merge({
				direction: "received",
				delivered: true,
				deliveredDBStored: true
			})));
	},
	messageReceivedDBStored: (state, action) => {
		let {messageId} = action;

		return state.update(markMessageDbStored(messageId));
	},
	messageSent: (state, action) => {
		let {message} = action;

		return state
			.update(addMessage(message.recipient, message.merge({
				direction: "sent"
			})));
	},
	messageSentDBStored: (state, action) => {
		let {messageId} = action;

		return state.update(markMessageDbStored(messageId));
	},
	messageDelivered: (state, action) => {
		// Defined as at least one device of the recipient having received the message, and acknowledged as such
		let {messageId} = action;

		return state.mergeIn(["messages", messageId], {
			delivered: true,
			queuedForOfflineDelivery: false
		});
	},
	messageDeliveredDBStored: (state, action) => {
		let {messageId} = action;

		return state.setIn(["messages", messageId, "deliveredDBStored"], true);
	},
	messageQueuedForOfflineDelivery: (state, action) => {
		let {messageId} = action;

		return state.setIn(["messages", messageId, "queuedForOfflineDelivery"], true);
	},
	friendAdded: (state, action) => {
		let {friend} = action;
		
		return state.updateIn(["friends"], (friends) => friends.set(friend.username, friend));
	},
	friendAddedDBStored: (state, action) => {
		let {username} = action;

		return state.setIn(["friends", username, "dbStored"], true);
	},
	friendRemoved: (state, action) => {
		let {username} = action;

		return state.updateIn(["friends"], (friends) => friends.delete(username));
	},
	friendRemovedDBStored: (state, action) => {
		/* Do nothing for now, there's not really a reasonable way to represent this in the current state tree structure. */
		
		return state;
	},
	friendRequestReceived: (state, action) => {
		let {request} = action;

		return state.updateIn(["friendRequests"], (requests) => requests.set(request.username, request));
	},
	friendRequestWithdrawn: (state, action) => {
		let {username} = action;

		return state.update(removeFriendRequest(username));
	},
	friendRequestAccepted: (state, action) => {
		let {username} = action;

		/* NOTE: This *only* handles the `friendRequests` entry, not the addition to the friends list. That's a separate `friendAdded` action. */

		return state.update(removeFriendRequest(username));
	},
	friendRequestedAcceptedDBStored: (state, action) => {
		/* Nothing? */
		return state;
	},
	friendRequestRejected: (state, action) => {
		let {username} = action;

		return state.update(removeFriendRequest(username));
	},
	friendRequestRejectedDBStored: (state, action) => {
		/* Nothing? */
		return state;
	}
};