"use strict";

const immutable = require("immutable");
const mapObj = require("map-obj");
const assureArray = require("assure-array");
const defaultValue = require("default-value");

const createLogic = require("./create-logic");
const scopedKey = require("./modules/scoped-key");
const unscopedKey = require("./modules/unscoped-key");

function reduceToObject(items, mapper = (item) => item) {
	return items.reduce((object, item) => {
		return Object.assign(object, mapper(item));
	}, {});
}

function reduceToArray(items, mapper = (item) => item) {
	return items.reduce((array, item) => {
		return array.concat(assureArray(mapper(item)));
	}, []);
}

function shallowClone(object) {
	return Object.assign({}, object);
}

function resolveMixins(module) {
	if (module.mixins == null) {
		return {
			state: defaultObject(module.state),
			stateTransformations: defaultObject(module.stateTransformations),
			logicHandlers: defaultObject(module.logicHandlers)
		};
	} else {
		let moduleClone = {
			state: shallowClone(defaultObject(module.state)),
			stateTransformations: shallowClone(defaultObject(module.stateTransformations)),
			logicHandlers: shallowClone(defaultObject(module.logicHandlers))
		};

		return module.mixins.reduce((resolved, mixin) => {
			/* NOTE: Mutation occurs here, but we pass in a (shallow) clone as initial value. */
			Object.assign(resolved.state, mixin.state);
			Object.assign(resolved.stateTransformations, mixin.stateTransformations);
			Object.assign(resolved.logicHandlers, mixin.logicHandlers);
			return resolved;
		}, moduleClone);
	}
}

function resolveTopLevelModule(module) {
	if (module.name == null) {
		/* FIXME: Allow this in the future? */
		throw new Error("Module is missing a `name` property");
	}

	let {state, stateTransformations, logicHandlers} = resolveMixins(module);

	return {
		name: module.name,
		state: immutable.Record(state)(),
		stateTransformations: stateTransformations,
		logicHandlers: logicHandlers
	};
}

function defaultObject(value) {
	return defaultValue(value, {});
}

/* TODO: Maybe eventually add support for dependency specifications, so that modules can safely refer to the scope of other modules, without depending on hard-coded prefixes in identifiers? */

module.exports = function loadModules(modules, unscoped) {
	let resolvedModules = modules.map((module) => resolveTopLevelModule(module));

	let allModules = resolvedModules.concat([unscoped]);
	let unscopedModules = allModules.filter((module) => (module.name == null));
	let scopedModules = allModules.filter((module) => (module.name != null));

	let stateModel = reduceToObject(unscopedModules, (module) => module.state);
	let stateModelModules = reduceToObject(scopedModules, (module) => ({[module.name]: module.state}));

	stateModel.modules = immutable.Record(stateModelModules)();
	let State = immutable.Record(stateModel);

	/* FIXME: Add support for multiple state transformations (eg. the default transformation and a module-provided one) */
	let scopedStateTransformations = reduceToObject(resolvedModules, (module) => {
		return mapObj(module.stateTransformations, (key, reducer) => {
			return [scopedKey(module.name, key), (state, action) => {
				let scopedState = reducer(state.modules[module.name], action);
				return state.setIn(["modules", module.name], scopedState);
			}];
		});
	});

	let stateTransformations = Object.assign(scopedStateTransformations, unscoped.stateTransformations);

	/* FIXME: Flatten array instead */
	let scopedLogicHandlers = reduceToArray(resolvedModules, (module) => {
		return Object.keys(module.logicHandlers).map((actionName) => {
			let options = module.logicHandlers[actionName];

			return createLogic({
				type: scopedKey(module.name, actionName),
				errorType: options.errorType, /* This is not scoped here, because it's handled in the actionMapper on dispatch */
				actionMapper: (action) => {
					return Object.assign({}, action, {
						type: scopedKey(module.name, action.type)
					});
				},
				actionUnmapper: (action) => {
					return Object.assign({}, action, {
						type: unscopedKey(module.name, action.type)
					});
				},
				dataMapper: (data) => {
					return Object.assign({}, data, {
						globalState: data.getState(),
						state: data.getState().modules[module.name]
					});
				},
				validate: options.validate,
				transform: options.transform,
				process: options.process,
				success: options.success,
				failure: options.failure
			});
		});
	});

	let unscopedLogicHandlers = Object.keys(defaultValue(unscoped.logicHandlers, {})).map((actionName) => {
		let options = unscoped.logicHandlers[actionName];

		return createLogic(Object.assign({}, options, {
			type: actionName
		}));
	});

	let logicHandlers = scopedLogicHandlers.concat(unscopedLogicHandlers);

	console.log("All registered transformations:", Object.keys(stateTransformations));
	console.log("All registered logic handlers:", logicHandlers.map((handler) => handler.type));
	
	return {
		State, stateTransformations, logicHandlers
	};
}