"use strict";

const immutable = require("immutable");
const redux = require("redux");

const match = require("./match");

let Friend = immutable.Record({
	username: null, // string
	displayName: null, // string
	status: null, // {"online", "away", "offline"}
	statusMessage: null, // string
	isTyping: false, // boolean
	pausedTyping: false, // boolean
	messageIds: immutable.List(), // (messages) id
	mutual: false
});

let FriendRequest = immutable.Record({
	username: null, // string
	message: null, // string
});

let UiFormState = immutable.Record({
	processing: false, // boolean
	errorMessage: null, // string
	/* QUESTION: How to handle form clearing? We don't really want to synchronize the form state to the state tree all the time, yet otherwise it may not be possible to have the form listen to the state tree changes, because react-redux assumes that everything you connect between a component and a state tree is either a data entry in the state tree, or an action creator. */
})

let UiAddFriend = immutable.Record({
	
})

let State = immutable.Record({
	connected: false, // boolean
	selectedTab: null, // string
	selectedTabType: null, // {"conversation", "meta"}
	username: null, // string
	displayName: null, // string
	displayNameDBStored: true, // string
	status: null, // {"online", "away", "offline"}
	statusMessage: null, // string
	isTypingTo: immutable.Set(), // (friends) username
	pausedTypingTo: immutable.Set(), // (friends) username
	friends: immutable.Map(), // string [username] -> Friend
	friendRequests: immutable.Map(), // string [username] -> FriendRequest
	messages: immutable.Map(), // string [id] -> Message
	uiAddFriend: null, // UiAddFriend
});

let Message = immutable.Record({
	id: null, // string
	direction: null, // {"sent", "received"}
	sender: null, // null | (friends) username
	recipient: null, // null | (friends) username
	timestamp: null, // number
	body: null, // string
	dbStored: false, // boolean
	delivered: false, // boolean
	deliveredDBStored: false, // boolean
	queuedForOfflineDelivery: false, // boolean
});

function addMessage(friendUsername, message) {
	return function (state) {
		return state
			.updateIn(["messages"], (messages) => messages.set(message.id, message))
			.updateIn(["friends", friendUsername, "messageIds"], (ids) => ids.push(message.id));
	}
}

function markMessageDbStored(messageId) {
	return function (state) {
		return state.setIn(["messages", messageId, "dbStored"], true);
	}
}

function removeFriendRequest(username) {
	return function (state) {
		return state.updateIn(["friendRequests"], (requests) => requests.delete(username));
	}
}

function clearFriendState(username) {
	return function (state) {
		return state.mergeIn(["friends", username], {
			isTyping: false,
			pausedTyping: false
		});
	}
}

function validateStatus(status) {
	if (!["online", "away", "offline"].includes(status)) {
		throw new Error(`Invalid status: ${status}`);
	} else {
		return status;
	}
}

function reducer(state, action) {
	return match(action.type, {
		selectedConversationTab: () => {
			let {username} = action;

			return state.merge({
				selectedTabType: "conversation",
				selectedTab: username
			});
		},
		selectedAddFriendTab: () => {
			return state.merge({
				selectedTabType: "meta",
				selectedTab: "addFriend"
			});
		},
		connected: () => {
			let {username, displayName, statusMessage, status, friends} = action;

			return state.merge({
				connected: true,
				username: username,
				displayName: displayName,
				status: validateStatus(status),
				statusMessage: statusMessage,
				friends: immutable.Map(friends.map((friend) => {
					return [friend.username, friend];
				}))
			});
		},
		disconnected: () => {
			return state.merge({
				connected: false
			});
		},
		userChangedStatus: () => {
			let {status} = action;

			return state.set("status", validateStatus(status));
		},
		userChangedStatusMessage: () => {
			let {statusMessage} = action;

			return state.set("statusMessage", statusMessage);
		},
		userChangedDisplayName: () => {
			let {displayName} = action;

			return state.merge({
				displayName: displayName,
				displayNameDBStored: false
			});
		},
		userChangedDisplayNameDBStored: () => {
			return state.set("displayNameDBStored", true);
		},
		userStartedTyping: () => {
			let {recipient} = action;

			return state.updateIn(["isTypingTo"], (friends) => {
				return friends.add(recipient);
			});
		},
		userPausedTyping: () => {
			let {recipient} = action;

			return state.updateIn(["pausedTypingTo"], (friends) => {
				return friends.add(recipient);
			});
		},
		userResumedTyping: () => {
			let {recipient} = action;

			return state.updateIn(["pausedTypingTo"], (friends) => {
				return friends.delete(recipient);
			});
		},
		userStoppedTyping: () => {
			let {recipient} = action;

			return state.updateIn(["isTypingTo"], (friends) => {
				return friends.delete(recipient);
			});
		},
		friendChangedStatus: () => {
			let {username, status} = action;

			/* TODO: Figure out a way to express the branching below in a functional, expression-based manner. */
			let state_ = state.setIn(["friends", username, "status"], status);

			if (status === "offline") {
				return state_.update(clearFriendState(username));
			} else {
				return state_;
			}
		},
		friendChangedStatusMessage: () => {
			let {statusMessage} = action;

			return state.setIn(["friends", username, "statusMessage"], statusMessage);
		},
		friendChangedDisplayName: () => {
			let {username, displayName} = action;

			return state.setIn(["friends", username, "displayName"], displayName);
		},
		friendStartedTyping: () => {
			return state.setIn(["friends", username, "isTyping"], true);
		},
		friendPausedTyping: () => {
			return state.setIn(["friends", username, "pausedTyping"], true);
		},
		friendResumedTyping: () => {
			return state.setIn(["friends", username, "pausedTyping"], false);
		},
		friendStoppedTyping: () => {
			return state.setIn(["friends", username, "isTyping"], false);
		},
		messageReceived: () => {
			let {message} = action;

			return state
				.update(addMessage(message.sender, message.merge({
					direction: "received",
					delivered: true,
					deliveredDBStored: true
				})));
		},
		messageReceivedDBStored: () => {
			let {messageId} = action;

			return state.update(markMessageDbStored(messageId));
		},
		messageSent: () => {
			let {message} = action;

			return state
				.update(addMessage(message.recipient, message.merge({
					direction: "sent"
				})));
		},
		messageSentDBStored: () => {
			let {messageId} = action;

			return state.update(markMessageDbStored(messageId));
		},
		messageDelivered: () => {
			// Defined as at least one device of the recipient having received the message, and acknowledged as such
			let {messageId} = action;

			return state.mergeIn(["messages", messageId], {
				delivered: true,
				queuedForOfflineDelivery: false
			});
		},
		messageDeliveredDBStored: () => {
			let {messageId} = action;

			return state.setIn(["messages", messageId, "deliveredDBStored"], true);
		},
		messageQueuedForOfflineDelivery: () => {
			let {messageId} = action;

			return state.setIn(["messages", messageId, "queuedForOfflineDelivery"], true);
		},
		friendAdded: () => {
			let {friend} = action;
			
			return state.updateIn(["friends"], (friends) => friends.set(friend.username, friend));
		},
		friendAddedDBStored: () => {
			let {username} = action;

			return state.setIn(["friends", username, "dbStored"], true);
		},
		friendRemoved: () => {
			let {username} = action;

			return state.updateIn(["friends"], (friends) => friends.delete(username));
		},
		friendRemovedDBStored: () => {
			/* Do nothing for now, there's not really a reasonable way to represent this in the current state tree structure. */
			
			return state;
		},
		friendRequestReceived: () => {
			let {request} = action;

			return state.updateIn(["friendRequests"], (requests) => requests.set(request.username, request));
		},
		friendRequestWithdrawn: () => {
			let {username} = action;

			return state.update(removeFriendRequest(username));
		},
		friendRequestAccepted: () => {
			let {username} = action;

			/* NOTE: This *only* handles the `friendRequests` entry, not the addition to the friends list. That's a separate `friendAdded` action. */

			return state.update(removeFriendRequest(username));
		},
		friendRequestedAcceptedDBStored: () => {
			/* Nothing? */
			return state;
		},
		friendRequestRejected: () => {
			let {username} = action;

			return state.update(removeFriendRequest(username));
		},
		friendRequestRejectedDBStored: () => {
			/* Nothing? */
			return state;
		},
		_: () => {
			return state;
		}
	});
}

module.exports = {
	State, Message, Friend, FriendRequest,
	createStore: function createStore(initialState = State()) {
		let devTools;

		if (window != null && window.__REDUX_DEVTOOLS_EXTENSION__ != null) {
			devTools = window.__REDUX_DEVTOOLS_EXTENSION__();
		}

		return redux.createStore(reducer, initialState, devTools);
	}
};